using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace ConvexHulls
{
    class ChanHull : IConvexHull
    {
        /// <summary>
        /// Calculates the convex hull.
        /// </summary>
        public Point[] CalculateHull(Point[] points)
        {
            // trivial cases
            if (points.Length <= 2)
                return points;

            // run partial hull with different sizes m for point sets until the hull is complete
            bool end = false;
            Point[] lhull = null;
            for (int i = 1; !end; i++)
            {
                int m = (int)Math.Pow(2, Math.Pow(2, i));
                lhull = partialHull(points, m);
                if (lhull != null)
                    end = true;
            } while (!end) ;
            return lhull;
        }

        /// <summary>
        /// Computes if Point p1 or p2 is the better next hull point
        /// </summary>
        /// <param name="p1">Point 1</param>
        /// <param name="p2">Point 2</param>
        /// <param name="lastPoint">last hull point</param>
        /// <returns>1 if p1 is better, -1 if p2 is better, 0 if both are equally good</returns>
        private int compare(Point p1, Point p2, Point lastPoint)
        {
            switch (pLeftOrRightOfVector(p1, lastPoint, p2))
            {
                case 1: return 1;
                case -1: return -1;
                case 0: //on the line - take point further away
                    double val1 = Math.Abs(lastPoint.X - p1.X) + Math.Abs(lastPoint.Y - p1.Y);
                    double val2 = Math.Abs(lastPoint.X - p2.X) + Math.Abs(lastPoint.Y - p2.Y);
                    if (val1 > val2) return 1;
                    else if (val1 < val2) return -1;
                    else return 0;
                default: return 0;
            }
        }

        private Point[] partialHull(Point[] points, int m)
        {
            // partitioning: divide points in n/m point sets
            // the resulting point sets are randomly distributed if the points are as well
            if (m > points.Length) m = points.Length;
            int r = (int)Math.Ceiling((double)points.Length / m);
            Point[][] psets = new Point[r][];
            int i = 0;
            for (int j = 0; j < psets.GetLength(0); j++)
            {
                int n = m;
                // last point set gets at most m points
                if (j == psets.GetLength(0) - 1)
                    if (points.Length % m != 0) n = points.Length % m;
                psets[j] = new Point[n];
                for (int k = 0; k < n; k++)
                {
                    psets[j][k] = points[i++];
                }
            }

            // Graham's Scan for the partitioned point sets
            GrahamScan gscan = new GrahamScan();
            Point[][] hulls = new Point[r][];
            for (int j = 0; j < psets.GetLength(0); j++)
            {
                hulls[j] = gscan.CalculateHull(psets[j]);
            }
            Point[] p = new Point[m + 1];     //hull points

            //p[0] = the rightmost point of P
            p[0] = new Point();
            foreach (Point pnt in points) if (pnt.X > p[0].X || (pnt.X == p[0].X && pnt.Y < p[0].Y)) p[0] = pnt;

            for (int k = 0; k < m; k++)    //all hull points
            {
                Point[] q = new Point[r];
                for (i = 0; i < r; i++) //all point sets
                {
                    //in this set find the best next hull point by binary search
                    Point next = new Point();
                    if (hulls[i].Length == 1) next = hulls[i][0];
                    else if (compare(hulls[i][0], hulls[i][1], p[k]) > 0
                        && compare(hulls[i][0], hulls[i][hulls[i].Length - 1], p[k]) > 0)
                        next = hulls[i][0]; //best point found
                    else if (compare(hulls[i][0], hulls[i][1], p[k]) < 0) //go from left to right
                    {
                        int startIndex = 0;
                        int endIndex = hulls[i].Length - 1;
                        int midIndex;
                        bool found = false;
                        while (!found)
                        {
                            if (endIndex == startIndex)
                            {
                                found = true;
                                next = hulls[i][startIndex];
                            }
                            else if (endIndex == startIndex + 1)
                            {
                                if (compare(hulls[i][startIndex], hulls[i][endIndex], p[k]) >= 0) next = hulls[i][startIndex];
                                else next = hulls[i][endIndex];
                                found = true;
                            }
                            else
                            {
                                midIndex = (endIndex + startIndex) / 2;
                                if (compare(hulls[i][startIndex], hulls[i][midIndex], p[k]) >= 0)
                                    endIndex = midIndex - 1;
                                else if (compare(hulls[i][midIndex - 1], hulls[i][midIndex], p[k]) >= 0)
                                    endIndex = midIndex - 1;
                                else startIndex = midIndex;
                            }
                        }
                    }
                    else //go from right to left
                    {
                        int startIndex = 0;
                        int endIndex = hulls[i].Length - 1;
                        int midIndex;
                        bool found = false;
                        while (!found)
                        {
                            if (endIndex == startIndex)
                            {
                                found = true;
                                next = hulls[i][startIndex];
                            }
                            else if (endIndex == startIndex + 1)
                            {
                                if (compare(hulls[i][startIndex], hulls[i][endIndex], p[k]) >= 0) next = hulls[i][startIndex];
                                else next = hulls[i][endIndex];
                                found = true;
                            }
                            else
                            {
                                midIndex = (endIndex + startIndex) / 2;
                                if (compare(hulls[i][endIndex], hulls[i][midIndex], p[k]) >= 0)
                                    startIndex = midIndex + 1;
                                else if (compare(hulls[i][midIndex + 1], hulls[i][midIndex], p[k]) >= 0)
                                    endIndex = midIndex + 1;
                                else endIndex = midIndex;
                            }
                        }
                    }
                    q[i] = next;
                }
                //find the best point among the best points of each point set
                Point nextHullPoint = q[0];
                foreach (Point pnt in q)
                {
                    switch (pLeftOrRightOfVector(pnt, p[k], nextHullPoint))
                    {
                        case 1: nextHullPoint = pnt; break;
                        case 0: //on the line - take point further away
                            if (Math.Abs(p[k].X - pnt.X) + Math.Abs(p[k].Y - pnt.Y)
                                > Math.Abs(p[k].X - nextHullPoint.X) + Math.Abs(p[k].Y - nextHullPoint.Y))
                                nextHullPoint = pnt;
                            break;
                    }
                }
                p[k + 1] = nextHullPoint;
                if (p[k + 1] == p[0]) // the hull is complete
                {
                    Point[] result = new Point[k + 1];
                    for (i = 0; i <= k; i++) result[i] = p[i];
                    return result;
                }
            }
            return null; // hull not complete
        }

        private Point[] partialHullLinearSearch(Point[] points, int m)
        {
            // partitioning: divide points in n/m point sets
            // the resulting point sets are randomly distributed if the points are as well
            if (m > points.Length) m = points.Length;
            int r = (int)Math.Ceiling((double)points.Length / (double)m);
            Point[][] psets = new Point[r][];
            int i = 0;
            for (int j = 0; j < psets.GetLength(0); j++)
            {
                int n = m;
                // last point set gets at most m points
                if (j == psets.GetLength(0) - 1)
                    if (points.Length % m != 0) n = points.Length % m;
                psets[j] = new Point[n];
                for (int k = 0; k < n; k++)
                {
                    psets[j][k] = points[i++];
                }
            }

            // Graham's Scan for the partitioned point sets
            GrahamScan gscan = new GrahamScan();
            Point[][] hulls = new Point[r][];
            for (int j = 0; j < psets.GetLength(0); j++)
            {
                hulls[j] = gscan.CalculateHull(psets[j]);
            }
            Point[] p = new Point[m + 1];     //hull points

            //p[0] = the rightmost point of P
            p[0] = new Point();
            foreach (Point pnt in points) if (pnt.X > p[0].X || (pnt.X == p[0].X && pnt.Y < p[0].Y)) p[0] = pnt;
            
            for (int k = 0; k < m; k++)    //all hull points
            {
                Point[] q = new Point[r];
                for (i = 0; i < r; i++) //all point sets
                {
                    //in this set find the best next hull point by linear search
                    
                    Point next = hulls[i][0];
                    foreach (Point pnt in hulls[i])
                    {
                        if (compare(pnt, next, p[k]) > 0) next = pnt;
                    }
                    /*
                    Point next = new Point();
                    if (hulls[i].Length == 1) next = hulls[i][0];
                    else if (compare(hulls[i][0], hulls[i][1], p[k]) < 0) //go from left to right
                    {
                        next = hulls[i][0];
                        for (int j = 1; j < hulls[i].Length; j++)
                        {
                            if (compare(hulls[i][j], next, p[k]) > 0) next = hulls[i][j];
                            //else break;
                        }
                    }
                    else //go from right to left
                    {
                        next = hulls[i][hulls[i].Length - 1];
                        for (int j = hulls[i].Length - 2; j >= 0; j--)
                        {
                            if (compare(hulls[i][j], next, p[k]) > 0) next = hulls[i][j];
                            else break;
                        }
                    }*/
                    q[i] = next;
                }
                //find the best point among the best points of each point set
                Point nextHullPoint = q[0];
                foreach (Point pnt in q)
                {
                    if (compare(pnt, nextHullPoint, p[k]) > 0) nextHullPoint = pnt;
                }
                p[k + 1] = nextHullPoint;
                if (p[k + 1] == p[0]) // the hull is complete
                {
                    Point[] result = new Point[k + 1];
                    for (i = 0; i <= k; i++) result[i] = p[i];
                    return result;
                }
            }
            return null; // hull not complete
        }

        /// <summary>
        /// Checks wether a point is to the right or to the left of a line.
        /// </summary>
        /// <param name="p">The point</param>
        /// <param name="from">Point 1 of the vector that specifies the line</param>
        /// <param name="to">Point 2 of the vector that specifies the line</param>
        /// <returns>Returns -1 if the point p is on the left hand side of the vector from->to.
        /// Returns 1 if the point p is on the right hand side.
        /// Returns 0 if it is on the line.</returns>
        private int pLeftOrRightOfVector(Point p, Point from, Point to)
        {
            int val = (to.X - from.X) * (p.Y - from.Y) - (p.X - from.X) * (to.Y - from.Y);
            if (val > 0) return -1; //left
            if (val < 0) return 1;  //right
            return 0;
        }
    }
}
