using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace ConvexHulls
{
    /// <summary>
    /// Convex hull algorithm "Graham scan".
    /// Web: http://en.wikipedia.org/wiki/Gift_wrapping_algorithm
    /// </summary>
    class JarvisMarch : IConvexHull
    {
        /// <summary>
        /// Calculates the convex hull of a point set in O(n*h) where h denotes to the number of points in the hull.
        /// </summary>
        /// <param name="points">All the points</param>
        /// <returns>Points in the hull</returns>
        public Point[] CalculateHull(Point[] points)
        {
            if (points.Length <= 2)
            {
                Point[] copy = new Point[points.Length];
                points.CopyTo(copy, 0);
                return copy;
            }

            List<Point> hull = new List<Point>();

            //Find the leftmost point as the start point
            Point leftmost = points[0];
            foreach (Point p in points)
            {
                if (p.X < leftmost.X) leftmost = p;
                if (p.X == leftmost.X) if (p.Y < leftmost.Y) leftmost = p;
            }
            hull.Add(leftmost);

            int i = 0;
            Point next;
            do
            {
                next = hull[i]; //hull[i] equals a "not set" here
                //find a next point such that the current and the next point
                //define a line that is left to all other points
                foreach (Point p in points)
                {
                    if (p != hull[i])
                    {
                        if (next == hull[i]) next = p;
                        else switch(pLeftOrRightOfVector(p, next, hull[i]))
                        {
                            case -1: next = p; break; //this may not be, so pick the new point to define the line
                            case 0: //on the line - take point further away
                                if (Math.Abs(hull[i].X - p.X) + Math.Abs(hull[i].Y - p.Y)
                                    > Math.Abs(hull[i].X - next.X) + Math.Abs(hull[i].Y - next.Y))
                                    next = p;
                                break;
                        }
                    }
                }
                hull.Add(next);
                i++;
            } while (hull[0] != hull[i]);
            hull.RemoveAt(hull.Count - 1);
            return hull.ToArray();
        }

        /// <summary>
        /// Checks wether a point is to the right or to the left of a line.
        /// </summary>
        /// <param name="p">The point</param>
        /// <param name="from">Point 1 of the vector that specifies the line</param>
        /// <param name="to">Point 2 of the vector that specifies the line</param>
        /// <returns>Returns -1 if the point p is on the left hand side of the vector from->to.
        /// Returns 1 if the point p is on the right hand side.
        /// Returns 0 if it is on the line.</returns>
        private int pLeftOrRightOfVector(Point p, Point from, Point to)
        {
            int val = (to.X - from.X) * (p.Y - from.Y) - (p.X - from.X) * (to.Y - from.Y);
            if (val < 0) return -1;
            if (val > 0) return 1;
            return 0;     
        }
	}
}