using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace ConvexHulls
{
    /// <summary>
    /// Convex hull algorithm "Quickhull".
    /// </summary>
    class QuickHull : IConvexHull
    {
        /// <summary>
        /// Calculates the convex hull.
        /// </summary>
        public Point[] CalculateHull(Point[] points)
        {
            // trivial cases
            if (points.Length <= 2)
                return points;

            // find points with min and max x
            Point min = points[0];
            Point max = points[0];
            for (int i = 1; i < points.Length; i++)
            {
                int x = points[i].X;
                if (x < min.X || (x == min.X && points[i].Y < min.Y))
                    min = points[i];
                else if (x > max.X || (x == max.X && points[i].Y > max.Y))
                    max = points[i];
            }

            // divide point set in two subsets (left and right of line (min,max))
            List<Point> lset = new List<Point>();
            List<Point> rset = new List<Point>();
            Point lmax = Point.Empty;
            Point rmax = Point.Empty;
            int lmaxdist = 0;
            int rmaxdist = 0;
            foreach (Point p in points)
            {
                int dist = getDist(p, min, max);
                if (dist < 0)     // point is on the left
                {
                    dist = -dist;
                    lset.Add(p);
                    if (lmaxdist < dist)
                    {
                        lmaxdist = dist;
                        lmax = p;
                    }
                }
                else if(dist > 0) // point is on the right
                {
                    rset.Add(p);
                    if (rmaxdist < dist)
                    {
                        rmaxdist = dist;
                        rmax = p;
                    }
                }
            }

            // recursive calls
            List<Point> lhull = qhullRec(min, max, lmax, lset);
            List<Point> rhull = qhullRec(max, min, rmax, rset);

            // connect parts to hull
            List<Point> listhull = new List<Point>();
            listhull.Add(min);
            listhull.AddRange(rhull);
            listhull.Add(max);
            listhull.AddRange(lhull);

            return listhull.ToArray();
        }

        private List<Point> qhullRec(Point p1, Point p2, Point pmax, List<Point> points)
        {
            if (pmax == Point.Empty)
                return points;

            // create sub sets on left of (p1,pmax) and (pmax,p2)
            List<Point> set1 = new List<Point>();
            List<Point> set2 = new List<Point>();
            Point max1 = Point.Empty;
            Point max2 = Point.Empty;
            int maxdist1 = 0;
            int maxdist2 = 0;
            foreach (Point p in points)
            {
                int dist1 = getDist(p, p1, pmax);
                // point is on left of (p1,pmax)
                if (dist1 < 0)
                {
                    dist1 = -dist1;
                    set1.Add(p);
                    if (maxdist1 < dist1)
                    {
                        maxdist1 = dist1;
                        max1 = p;
                    }
                }
                else
                {
                    int dist2 = getDist(p, pmax, p2);
                    // point is on left of (pmax,p2)
                    if (dist2 < 0)
                    {
                        dist2 = -dist2;
                        set2.Add(p);
                        if (maxdist2 < dist2)
                        {
                            maxdist2 = dist2;
                            max2 = p;
                        }
                    }
                }
            }

            // recursive calls
            List<Point> hull1 = new List<Point>();
            if(max1 != Point.Empty)
                hull1 = qhullRec(p1, pmax, max1, set1);
            List<Point> hull2 = new List<Point>();
            if(max2 != Point.Empty)
                hull2 = qhullRec(pmax, p2, max2, set2);
            
            // connect hull parts
            hull2.Add(pmax);
            hull2.AddRange(hull1);

            return hull2;
        }

        /// <summary>
        /// Returns the distance of a point p to a vector (from,to).
        /// </summary>
        private int getDist(Point p, Point vfrom, Point vto)
        {
            return ((vto.X - vfrom.X) * (p.Y - vfrom.Y) - (p.X - vfrom.X) * (vto.Y - vfrom.Y));
        }
    }
}
