package algorithms.priorityqueues;

import java.util.Hashtable;
import java.util.Map;
import java.util.Set;

public class LeftistTree<K,V> implements PriorityQueue<K,V> {
	
	LeftistTreeNode<K,V> root = null;
	Map<K,LeftistTreeNode<K,V>> dataAccess = new Hashtable<K,LeftistTreeNode<K,V>>();
	
	
	public LeftistTree() { }
	public LeftistTree(LeftistTreeNode<K,V> root){
		this.root = root;
	}
	public LeftistTree(V[] values, K[] keys, int[] priorities) throws IllegalArgumentException{
		if(values.length!=priorities.length || values.length!=keys.length)
			throw new IllegalArgumentException("Argument arrays must have the same dimension");
		
		for(int i=0; i<values.length; i++){
			LeftistTreeNode<K,V> node = new LeftistTreeNode<K,V>(values[i], keys[i], priorities[i]);
			insert(node);
		}
	}
	
	public Set<K> getKeySet() {
		return dataAccess.keySet();
	}
	
	public LeftistTreeNode<K,V> getNode(K key) {
		return dataAccess.get(key);
	}
	
	public int size(){
		return size(root);
	}
	
	private int size(LeftistTreeNode<K,V> node){
		if(null == node)
			return 0;
		else
			return size(node.getLeftChild()) + size(node.getRightChild()) + 1;
			
	}
	
	boolean correctDist;
	public boolean checkQueueIntegrity(){
		correctDist = true;
		checkDistances(root);
		return checkQueueIntegrity(root);
	}
	
	private boolean checkQueueIntegrity(LeftistTreeNode<K,V> node){
		if(null == node)
			return true;
		if((null!=node.getLeftChild() && node.getLeftChild().getPriority()<node.getPriority()) ||
		 (null!=node.getRightChild() && node.getRightChild().getPriority()<node.getPriority()) ||
		 (null!=node.getRightChild() && null!=node.getLeftChild() && node.getLeftChild().getDistance()<node.getRightChild().getDistance())){
			return false;
		}
		return checkQueueIntegrity(node.getLeftChild()) && checkQueueIntegrity(node.getRightChild());
	}
	
	private int checkDistances(LeftistTreeNode<K,V> node){
		if(null == node)
			return 0;
		int dist = 0;
		if(null!=node.getLeftChild() && null!=node.getRightChild())
			dist = Math.min(checkDistances(node.getLeftChild()), checkDistances(node.getRightChild()));
		
		if(dist != node.getDistance())
			correctDist = false;
		
		return dist+1;
	}
	
	public void decreasePriority(PriorityQueueNode<K,V> node, int newValue) {
		LeftistTreeNode<K,V> n = (LeftistTreeNode<K,V>)node;
		delete(n);
		n.setPriority(newValue);
		insert(n);
	}

	private boolean contains(LeftistTreeNode<K,V> base, LeftistTreeNode<K,V> node){
		if(base == null)
			return false;
		if(base == node)
			return true;
		else
			return contains(base.getLeftChild(),node) || contains(base.getRightChild(), node);
	}
	
	private boolean hasSameSon(LeftistTreeNode<K,V> node){
		if(node == null)
			return false;
		if(node == node.getLeftChild() || node == node.getRightChild())
			return true;
		return hasSameSon(node.getLeftChild()) || hasSameSon(node.getRightChild());
	}
	
	public boolean delete(PriorityQueueNode<K,V> node) {
		if(null == node)
			return false;
		
		LeftistTreeNode<K,V> n = (LeftistTreeNode<K,V>)node;
		LeftistTreeNode<K,V> merged = merge(n.getLeftChild(), n.getRightChild());
		
		if(n == root)
			root = merged;
		
		if(null != n.getParent()){
			if( n.isLeftChild() ){
				n.getParent().setLeftChild(merged);
				if(null == merged)
					n.getParent().swapChildren();
			}
			else
				n.getParent().setRightChild(merged);
		}
		
		if(null != merged){
			merged.setParent(n.getParent());
			merged.updateDistance();
		}
		
		LeftistTreeNode<K,V> parent = n.getParent();
		// go along path to root and update distances
		while(null != parent){
			LeftistTreeNode<K,V> left = parent.getLeftChild();
			LeftistTreeNode<K,V> right = parent.getRightChild();
			
			if(null==left || (null!=left && null!=right && left.getDistance()<right.getDistance())){
				parent.swapChildren();
			}
			parent.updateDistance();
			parent = parent.getParent();
		}
		
		n.setParent(null);
		n.setLeftChild(null);
		n.setRightChild(null);
		
		return true;
	}

	public PriorityQueueNode<K,V> deleteMinimum() {
		if(root == null)
			return null;
		
		LeftistTreeNode<K,V> min = root;
		this.root = merge(root.left, root.right);
		
		return min;
	}

	public PriorityQueueNode<K,V> getMinimum() {
		return root;
	}

	public void insert(PriorityQueueNode<K,V> x) {
		LeftistTreeNode<K,V> x1 = (LeftistTreeNode<K,V>)x;
		this.root = merge(this.root, x1);
		dataAccess.put(x1.getKey(), x1);
	}

	@SuppressWarnings("unchecked")
	public void merge(PriorityQueue<K,V> q2) {
		LeftistTree<K,V> lt = (LeftistTree<K,V>)q2;
		merge(lt.root);
		// add lt hashtable to this hashtable
		dataAccess.putAll(lt.dataAccess);
	}
	
	private void merge(LeftistTreeNode<K,V> q2) {
		this.root = merge(this.root, q2);
	}
	
	private LeftistTreeNode<K,V> merge(LeftistTreeNode<K,V> a, LeftistTreeNode<K,V> b){
		if(null == a)
			return b;
		if(null == b)
			return a;
		
		if(a.getPriority() > b.getPriority()){
			LeftistTreeNode<K,V> tmp = a;
			a = b;
			b = tmp;
		}
		
		LeftistTreeNode<K,V> merged = merge(a.getRightChild(), b);
		a.setRightChild(merged);
		merged.setParent(a);
		if(a.getLeftChild()==null || a.getRightChild()!=null && a.getLeftChild()!=null && a.getRightChild().getDistance()>a.getLeftChild().getDistance()){
			a.swapChildren();
		}
		if(a.getRightChild() == null)
			a.setDistance(0);
		else
			a.setDistance(a.getRightChild().getDistance()+1);
		
		return a;
	}
}
